/* 
 *
 * robotAI
 * Version 1.1
 * July 26 2008
 * Released under GNU GPL
 *
 * robotAI ships freight. with trucks
 *
 */

import("pathfinder.road", "RoadPathFinder", 3);


/*
 *
 * Class - Line
 *
 */

import("pathfinder.road", "RoadPathFinder", 3);

class Line
{
	source = null;
	destination = null;
	station1 = null;
	station1tile = null;
	station1_crowded = false;
	station2 = null;
	station2tile = null;
	station2_crowded = false;
	depottile = null;
	Vehicles = null;
	cargo = null;
	station_site_count = 0;
	vps_target = 5;
	group_id = 0;

	constructor(source, destination, cargo)
	{
		this.source = source;
		this.destination = destination;
		this.cargo = cargo;
		this.Vehicles = [];
		this.station1_crowded = false;
		this.station2_crowded = false;
		this.group_id = 0;
	}

	function InitializeRoute(whoami);
	function BuildStation(target, siteType);
	function FindFacingRoadTile(tile);
	function FindFacingEmptyTile(tile);
	function FindSite(target, siteType, radius);
	function GetCargoRating();
	function AddAdditionalVehicle();
	function ExpandStation(number);
	function ExpandStations();
	function GetNumVehiclesUnprofitable();
	function GetListVehiclesUnprofitable();
	function CullVehicles();
	function SearchAndDestroy(vehicle_id);
}

/*
 * function Line.InitializeRoute;
 */

function Line::InitializeRoute(whoami)
{
  // AILog.Info("Line: initializing route");      

  // build the stations

	local station1Site = this.BuildStation(source, 1);
	local station2Site = this.BuildStation(destination, 2);

	if (station1Site[0] == -1 || station1Site[1] == 1)
	{
		AILog.Info("Line: Could not build station 1!");
		return false;
	}

	if (station2Site[0] == -1 || station2Site[1] == 1)
	{
		AILog.Info("Line: Could not build station 2!");
		// clear the useless station 1
		AITile.DemolishTile(station1Site[0]);
		return false;
    }



  // find the route       

	local pathfinder = RoadPathFinder();

	pathfinder.cost.no_existing_road = 400;
/*	pathfinder.cost.turn = 750;
	pathfinder.cost.coast = 500;
	pathfinder.cost.slope = 500;
	pathfinder.cost.max_tunnel_length = 30;
	pathfinder.cost.max_bridge_length = 30; */
	pathfinder.cost.bridge_per_tile = 1500;
	pathfinder.cost.tunnel_per_tile = 1300;


	pathfinder.InitializePath([station1Site[1]],[station2Site[1]]);

	local path = false;

	while (path == false)
	{

		path = pathfinder.FindPath(100);
		// AILog.Info("Line: Finding route...");                
		robotAI.Sleep(1);

	}

	if (path == null)
	{

		AILog.Info("Line: no path found!");
		// no route, so kill the station buildings
		AITile.DemolishTile(station1Site[0]);
		AITile.DemolishTile(station2Site[0]);	
		return false;

    }

	AILog.Info("Line: Found route");

	while (path != null)
	{

		local par = path.GetParent();
		if (path.GetParent() != null)
		{

			if (AIMap.DistanceManhattan(path.GetTile(), par.GetTile()) == 1)
			{

// it is possible part of the path is blocked, so if building a tile fails
// check for that

				// AILog.Info("Line: Building road...");
				local counter = 0
				while (!AIRoad.BuildRoad(path.GetTile(), par.GetTile()))
				{

					if (AIError.GetLastError() ==
						AIError.ERR_VEHICLE_IN_THE_WAY)
					{
						if (counter == 0)
						{
							AILog.Info
								("Line: Road vehicle in the way, retrying.");
							counter = 10;
						}
						counter--;
						AIController.Sleep(1);

					}

					else if (AIError.GetLastError() ==
						AIError.ERR_NOT_ENOUGH_CASH)

					{

						AILog.Warning("Line: Error: Not enough cash to" 
							+ " finish building route.");

		// so basically demolish buildings before returning false
		// the idea is to keep property maintenance down, dig?
						AITile.DemolishTile(station1Site[0]);
						AITile.DemolishTile(station2Site[0]);
						return false;

					}

// if that's not the case, maybe the tiles are already connected
					else if (AIRoad.AreRoadTilesConnected(path.GetTile(),
						par.GetTile()))
					{
						break;
					}

// if that's not the case, whine
					else
					{
						AILog.Warning("Line: Error building route");
						AITile.DemolishTile(station1Site[0]);
						AITile.DemolishTile(station2Site[0]);
						return false;
					}
				}
			}

			else
			{	// it's a bridge or a tunnel

				if (!AIBridge.IsBridgeTile(path.GetTile()) &&
					!AITunnel.IsTunnelTile(path.GetTile()))
				{

					if (AIRoad.IsRoadTile(path.GetTile()))
						AITile.DemolishTile(path.GetTile());
					if (AITunnel.GetOtherTunnelEnd(path.GetTile()) ==
						par.GetTile())
					{
	
						AILog.Info("Line: Building tunnel...");
						local counter = 0;
						while (!AITunnel.BuildTunnel
							(AIVehicle.VEHICLE_ROAD, path.GetTile()))
						{
	
							if (AIError.GetLastError() ==
								AIError.ERR_VEHICLE_IN_THE_WAY)
							{
								if (counter == 0)
								{
									AILog.Info("Line: Road vehicle in" 
										+ " the way, retrying.");
									counter = 10;
								}

								counter--;								
								AIController.Sleep(1);
	
							}
							else if (AIError.GetLastError() ==
								AIError.ERR_NOT_ENOUGH_CASH)
							{
	
								AILog.Warning("Line: Error: Not enough"
									+ " cash to finish building"
									+ "route.");

								AITile.DemolishTile(station1Site[0]);
								AITile.DemolishTile(station2Site[0]);
								return false;
	
							}
							else
							{
	
								AILog.Warning
									("Line: Error building route");
	
					// one of these days, i'm gonna write a Line.Kill function
					// to do this for me
	
								AITile.DemolishTile(station1Site[0]);
								AITile.DemolishTile(station2Site[0]);
								return false;
	
							}
						}
					}

					else
					{

						local bridges = AIBridgeList_Length
							(AIMap.DistanceManhattan(path.GetTile(), 
							par.GetTile() + 1));
						bridges.Valuate(AIBridge.GetMaxSpeed);
						bridges.Sort(AIAbstractList.SORT_BY_VALUE, false);
						AILog.Info("Line: building bridge...");

						local counter = 0;
						while (!AIBridge.BuildBridge
							(AIVehicle.VEHICLE_ROAD, bridges.Begin(), 
							path.GetTile(), par.GetTile()))
						{

							if (AIError.GetLastError() ==
								AIError.ERR_VEHICLE_IN_THE_WAY)
							{
								if (counter == 0)
								{
									AILog.Info("Line: Road vehicle in"
										+ " the way, retrying.");
									counter = 10;
								}
								counter--;
								AIController.Sleep(1);
							}

							else if (AIError.GetLastError() ==
								AIError.ERR_NOT_ENOUGH_CASH)
							{

								AILog.Warning("Line: Error: Not enough"
									+ " cash to finish building"
									+ " route.");
								AITile.DemolishTile(station1Site[0]);
								AITile.DemolishTile(station2Site[0]);
								return false;

							}

							else
							{

								AILog.Warning("Line: Error building" 
									+ " route");
								AITile.DemolishTile(station1Site[0]);
								AITile.DemolishTile(station2Site[0]);
								return false;

							}
						}
					}
				}
			}
		}

		path = par;

	}

  // now the path is built, we locate and build a depot

	AILog.Info("Line: Stations connected, building depot");
	local depotSite = this.FindSite(station1Site[0], 0, 10);
	local counter = 0;
	while (!AIRoad.BuildRoad(depotSite[0], depotSite[1]))
	{
		if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY)
		{
			if (counter == 0)
			{
				AILog.Info("Line: Road vehicle in the way, retrying.");
				counter = 10;
			}
			counter--;
			AIController.Sleep(1);
		}
		else
		{
			AILog.Warning("Line: Error build road at" + depotSite[0]);

			// in this case we've built the route so we may be 
			// screwed on property costs anyways
			AITile.DemolishTile(station1Site[0]);
			AITile.DemolishTile(station2Site[0]);
			return false;
		}
	}

	local counter = 0;

	while (!AIRoad.BuildRoadDepot(depotSite[0], depotSite[1]))
	{
		if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY)
		{
			if (counter == 0)
			{
				AILog.Info("Line: Road vehicle in the way, retrying.");
				counter = 10;
			}
			counter--;
			AIController.Sleep(1);
		}
		else
		{
			AILog.Warning("Line: Error build depot at" + depotSite[0]);
			AITile.DemolishTile(station1Site[0]);
			AITile.DemolishTile(station2Site[0]);
			return false;
		}
	}

	this.depottile = depotSite[0];
	this.station1tile = station1Site[0];
	this.station2tile = station2Site[0];
	this.station1 = AIStation.GetStationID(station1Site[0]);
	this.station2 = AIStation.GetStationID(station2Site[0]);

  // now there's two stations and a depot, so we can actually add a truck
  // find the best engine to use
	local possibleEngines = AIEngineList(AIVehicle.VEHICLE_ROAD);
	possibleEngines.Valuate(AIEngine.GetCargoType);
	possibleEngines.KeepValue(this.cargo);
	possibleEngines.Valuate(AIEngine.GetReliability);
	possibleEngines.Sort(AIList.SORT_BY_VALUE, false);
	possibleEngines.KeepTop(1);
	local engineID = possibleEngines.Begin();

  // build the vehicle
	local thisvehicle = AIVehicle.BuildVehicle(this.depottile, engineID);
	if (!AIVehicle.IsValidVehicle(thisvehicle))
	{
		AILog.Error("Line: Error building vehicle at " + this.depottile);
		AITile.DemolishTile(this.station1tile);
		AITile.DemolishTile(this.station2tile);
		AITile.DemolishTile(this.depottile);
		return false;
	}

	// give it some orders
	if (!AIOrder.AppendOrder(thisvehicle, this.station1tile, 
			AIOrder.AIOF_FULL_LOAD) 
		|| !AIOrder.AppendOrder(thisvehicle, this.depottile, 
			AIOrder.AIOF_NONE) 
		|| !AIOrder.AppendOrder(thisvehicle, this.station2tile, 
			AIOrder.AIOF_NONE))
	{
		AILog.Error("Line: error giving orders to vehicle " 
			+ thisvehicle);
		AIVehicle.SellVehicle(thisvehicle);
		AITile.DemolishTile(this.station1tile);
		AITile.DemolishTile(this.station2tile);
		AITile.DemolishTile(this.depottile);
		return false;
	}
	// start me up
	if (!AIVehicle.StartStopVehicle(thisvehicle))
	{
		AILog.Warning("Line: error starting vehicle " + thisvehicle);
		AIVehicle.SellVehicle(thisvehicle);
		AITile.DemolishTile(this.station1tile);
		AITile.DemolishTile(this.station2tile);
		AITile.DemolishTile(this.depottile);
		return false;
	}

	this.group_id = AIGroup.CreateGroup(AIVehicle.VEHICLE_ROAD);
	if (AIGroup.IsValidGroup(group_id))
	{
		AIGroup.SetName(group_id, ("line " + whoami));
		AIGroup.MoveVehicle(group_id, thisvehicle);
	}
	else 
	{
		group_id = -1;
	}
	this.Vehicles.append(thisvehicle);
	AILog.Info("Line: Line successfully built from " +
     AIStation.GetName(this.station1) + " to " +
     AIStation.GetName(this.station2));
	return true;
}





/*
 * function Line.BuildStation(target, siteType)
 *
 * target - industry ID of target stations
 * accepting - 1 for producer, 2 for accepting
 * builds a station to serve the target industry and returns it's location
 * checks for existing stations and extends them if possible
 * like Line.FindSite, returns [-1, -1] if fails
 */

function Line::BuildStation(target, siteType)
{

	local target_tile = 0;
	local site = [];

	if (AIIndustry.GetAmountOfStationsAround(target) > 0)
	{

		AILog.Info("BuildStation: Already is a station here");
		local coverage = AIStation.GetCoverageRadius
			(AIStation.STATION_TRUCK_STOP);
		local station_tiles = AITileList();

		if (siteType == 1)
		{
			station_tiles = AITileList_IndustryProducing(target, coverage);
		}
		else if (siteType == 2)
		{
			station_tiles = AITileList_IndustryAccepting(target, coverage);
		}

		station_tiles.Valuate(AIRoad.IsRoadStationTile);
		station_tiles.KeepValue(1);
		station_tiles.Valuate(AITile.GetOwner);
		station_tiles.KeepValue
			(AICompany.ResolveCompanyID(AICompany.MY_COMPANY));

		if (station_tiles.IsEmpty())
		{

			AILog.Warning("BuildStation: There's no station here!");
			target_tile = AIIndustry.GetLocation(target);
			site = this.FindSite(target_tile, siteType, 4);

		}
		else
		{
			AILog.Info("BuildStation: Found " + station_tiles.Count()
				 + " existing station sites.");

			// build a list of tiles that are adjacent to existing stations
			local candidate_sites = AITileList();

			foreach(thistile, val in station_tiles)
			{

				local templist = AITileList();
				templist.AddTile(thistile - AIMap.GetTileIndex(-1, -1));
				templist.AddTile(thistile - AIMap.GetTileIndex(-1, 0));
				templist.AddTile(thistile - AIMap.GetTileIndex(-1, 1));
				templist.AddTile(thistile - AIMap.GetTileIndex(0, -1));
				templist.AddTile(thistile - AIMap.GetTileIndex(0, 1));
				templist.AddTile(thistile - AIMap.GetTileIndex(1, -1));
				templist.AddTile(thistile - AIMap.GetTileIndex(1, 0));
				templist.AddTile(thistile - AIMap.GetTileIndex(1, 1));

				templist.Valuate(AIMap.IsValidTile);
				templist.KeepValue(1);
				templist.Valuate(AITile.GetSlope);
				templist.KeepValue(0);
				templist.Valuate(AITile.IsBuildable);
				templist.KeepValue(1);

			/*
				AILog.Info("BuildStation: " + templist.Count() + 
				" flat buildable tiles adjacent to station at " + thistile);
			*/

				candidate_sites.AddList(templist);
			}

			AILog.Info("BuildStation: " + candidate_sites.Count() +
				" candidate sites for station extension");

			if (candidate_sites.IsEmpty())
			{

				// no places to extend the station
				// so act like there's no station there at all
				AILog.Info("BuildStation: No candidate sites, doing it the "
					+ " old way");
				target_tile = AIIndustry.GetLocation(target);
				site = this.FindSite(target_tile, siteType, 4);

			}
			else
			{

				// otherwise try to figure out which tile is best
				local sites_with_roads = AITileList();
				sites_with_roads.AddList(candidate_sites);
				sites_with_roads.Valuate(this.FindFacingRoadTile);
				sites_with_roads.RemoveValue(-1);
				local sites_are_empty = AITileList();
				sites_are_empty.AddList(candidate_sites);
				sites_are_empty.Valuate(this.FindFacingEmptyTile);
				sites_are_empty.RemoveValue(-1);

				if (!sites_with_roads.IsEmpty())
				{

					sites_with_roads.Valuate(AIBase.RandItem);
					sites_with_roads.Sort(AIList.SORT_BY_VALUE, false);
					local site_tile = sites_with_roads.Begin();
					local site_front = this.FindFacingRoadTile(site_tile);
					site = [site_tile, site_front];
					AILog.Info("BuildStation: " + site[0] +
						" already faces a road, building there.");

				}
				else if (!sites_are_empty.IsEmpty())
				{

					// if no sites facing a road tile, try for empty tiles
					sites_are_empty.Valuate(AIBase.RandItem);
					sites_are_empty.Sort(AIList.SORT_BY_VALUE, false);
					local site_tile = sites_are_empty.Begin();
					local site_front = this.FindFacingEmptyTile(site_tile);
					site = [site_tile, site_front];
					AILog.Info("BuildStation: " + site[0] +
						" is an empty site, building there.");

				}
				else
				{

					// no valid sites to build the extension
					// so act like there's no station there at all
					AILog.Info("BuildStation: No candidate sites, doing it" 
						+ "the old way");
					target_tile = AIIndustry.GetLocation(target);
					site = this.FindSite(target_tile, siteType, 4);

				}
			}

		}
	}

	else
	{
		target_tile = AIIndustry.GetLocation(target);
		site = this.FindSite(target_tile, siteType, 4);
 	}

	local counter = 0;
	while (!AIRoad.BuildRoad(site[0], site[1]))
	{

		if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY)
		{
			if (counter == 0)
			{
				AILog.Info("BuildStation: Road vehicle in way, retrying.");
				counter = 10;
			}
			counter--;
			AIController.Sleep(1);
		}
		else
		{

			AILog.Error("BuildStation: Error building road at " + site[1]);
			return[-1, -1];

		}
	}

	if (!AIRoad.BuildRoadStation(site[0], site[1], true, false, true))
	{
		AILog.Error("BuildStation: Error building station at " + site[0]);
		AITile.DemolishTile(site[0]);

		return[-1, -1];
	}

	this.station_site_count = this.station_site_count + 1;
	return site;

}


/*
 * function Line.FindFacingRoadTile(tile)
 * returns the location of a random road tile adjacent to given tile
 * or -1 if there is none
 */

function Line::FindFacingRoadTile(tile)
{
  local
    adjacent = AITileList();
  adjacent.AddTile(tile - AIMap.GetTileIndex(1, 0));
  adjacent.AddTile(tile - AIMap.GetTileIndex(0, 1));
  adjacent.AddTile(tile - AIMap.GetTileIndex(-1, 0));
  adjacent.AddTile(tile - AIMap.GetTileIndex(0, -1));
  adjacent.Valuate(AIMap.IsValidTile);
  adjacent.KeepValue(1);
  adjacent.Valuate(AIRoad.IsRoadTile);
  adjacent.KeepValue(1);
  adjacent.Valuate(AITile.GetSlope);
  adjacent.KeepValue(0);
  adjacent.Valuate(AITile.GetHeight);
  adjacent.KeepValue(AITile.GetHeight(tile));
  if (adjacent.IsEmpty())
    {
      /* AILog.Warning("FindFacingRoadTile: no candidate road tiles at" + 
         tile); */
      return -1;
    }
  else
    {
      adjacent.Valuate(AIBase.RandItem);
      adjacent.Sort(AIList.SORT_BY_VALUE, false);
      return adjacent.Begin();
    }
}


/*
 * function Line.FindFacingEmptyTile(tile)
 * returns the location of a random buildable tile adjacent to given tile
 * or -1 if there is none
 */

function Line::FindFacingEmptyTile(tile)
{
  local
    adjacent = AITileList();
  adjacent.AddTile(tile - AIMap.GetTileIndex(1, 0));
  adjacent.AddTile(tile - AIMap.GetTileIndex(0, 1));
  adjacent.AddTile(tile - AIMap.GetTileIndex(-1, 0));
  adjacent.AddTile(tile - AIMap.GetTileIndex(0, -1));
  adjacent.Valuate(AIMap.IsValidTile);
  adjacent.KeepValue(1);
  adjacent.Valuate(AITile.IsBuildable);
  adjacent.KeepValue(1);
  adjacent.Valuate(AITile.GetSlope);
  adjacent.KeepValue(0);
  adjacent.Valuate(AITile.GetHeight);
  adjacent.KeepValue(AITile.GetHeight(tile));
  if (adjacent.IsEmpty())
    {
      /* AILog.Warning("FindFacingEmptyTile: no candidate road tiles at" + 
         tile); */
      return -1;
    }
  else
    {
      adjacent.Valuate(AIBase.RandItem);
      adjacent.Sort(AIList.SORT_BY_VALUE, false);
      return adjacent.Begin();
    }
}



/*
 * function Line.FindSite(target, siteType, radius)
 *
 *
 * siteType (int) - 0 depot, 1 producing station, 2 accepting station
 * radius (int) - the number of tile out from the target for the search 
 *      rectangle, ie 3 searches a 7x7 grid around the target 
 *      (this is not actually what radius means)
 *      
 * returns [site, front]
 * [-1, -1] means no suitable site was found
 *
 * this function really needs to be cleaned up methinks
 */

function Line::FindSite(target, siteType, radius)
{
  local
    possibilities = AITileList();
  local
    adjacent = AITileList();
  local
    attempt = 0;
  local
    count = 0;
  local
    foundtile = -1;
  while (count == 0)
    {
      AIController.Sleep(1);
      attempt = attempt + 1;	// this is not real problem solving
      if (attempt == 6)
	{
	  AILog.Warning("Line: Unable to locate site!");
	  return[-1, -1];
	}
      // AILog.Info("Line: Attempt " + attempt);
      local
	area = AITileList();
      area.AddRectangle(target - AIMap.GetTileIndex(radius, radius),
			target + AIMap.GetTileIndex(radius, radius));
      area.Valuate(AIMap.IsValidTile);
      area.KeepValue(1);

      if (siteType >> 0)
	{			// if it's a station we want empty tiles
	  area.Valuate(AITile.IsBuildable);
	  area.KeepValue(1);
	  local
	    coverage =
	    AIStation.GetCoverageRadius(AIStation.STATION_TRUCK_STOP);
	  if (siteType == 1)
	    {
	      // for a producing station we want sites that have cargo
	      area.Valuate(AITile.GetCargoProduction, this.cargo, 1, 1,
			   coverage);
	      area.RemoveValue(0);
	    }
	  else if (siteType == 2)
	    {
	      // for a recieving station we want sites that take cargo
	      area.Valuate(AITile.GetCargoAcceptance, this.cargo, 1, 1,
			   coverage);
	      area.RemoveValue(0);
	    }
	}
      else if (siteType == 0)
	{
	  area.Valuate(AIRoad.IsRoadTile);	// if it's a depot we want road tiles
	  area.KeepValue(1);
	}
      else
	{
	  AILog.Error("Line: Must specify depot or station!");
	  return[-1. - 1];
	}

      area.Valuate(AITile.GetSlope);
      area.KeepValue(0);
      // AILog.Info("Line: " + area.Count() + " tiles found.");       

      if (area.Count() == 0)
	{
	  return[-1, -1];
	}

      if (siteType >> 0)
	{			// for stations, choose a random tile
	  area.Valuate(AIBase.RandItem);
	}
      else
	{			// for depots, choose a random tile no farther than specified distance
	  area.Valuate(AITile.GetDistanceManhattanToTile, target);
	  area.KeepBelowValue(radius + 1);
	  area.Valuate(AIBase.RandItem);
	}
      area.Sort(AIList.SORT_BY_VALUE, true);

      foundtile = area.Begin();

      if (siteType == 0)
	{
	  AILog.Info("Line: " + area.Count() +
		     " possible tiles for depot to face.");
	  local
	    connector = RoadPathFinder();
	  connector.cost.no_existing_road = connector.cost.max_cost;
	  connector.cost.bridge_per_tile = connector.cost.max_cost;
	  connector.cost.tunnel_per_tile = connector.cost.max_cost;
	  local
	    connected = false;
	  while (area.HasNext())
	    {
	      // Need to make sure this is connected to the target            
	      AILog.Info("Line: verifying connection for tile " + foundtile);
	      connector.InitializePath([foundtile],[target]);
	      connected = connector.FindPath(100);

	      // checking of adjacent tiles is here to not get tricked by towns
	      // this probably means there's no need to check again later
	      local
		height = AITile.GetHeight(foundtile);
	      adjacent = AITileList();
	      adjacent.AddTile(foundtile - AIMap.GetTileIndex(1, 0));
	      adjacent.AddTile(foundtile - AIMap.GetTileIndex(0, 1));
	      adjacent.AddTile(foundtile - AIMap.GetTileIndex(-1, 0));
	      adjacent.AddTile(foundtile - AIMap.GetTileIndex(0, -1));

	      adjacent.Valuate(AIMap.IsValidTile);
	      adjacent.KeepValue(1);

	      adjacent.Valuate(AITile.GetHeight);
	      adjacent.KeepValue(height);
	      adjacent.Valuate(AITile.GetSlope);
	      adjacent.KeepValue(0);
	      adjacent.Valuate(AITile.IsBuildable);
	      adjacent.KeepValue(1);

	      if (adjacent.Count() == 0)
		{
		  connected = false;
		}

	      if (connected)
		{
		  break;
		}

	      foundtile = area.Next();
	      AIController.Sleep(1);
	    }
	  if (connected == false)
	    {
	      AILog.Info("Line: chosen site does not connect to target");
	      return[-1, -1];
	    }
	}

      // do we need to do this again for depots? probably not, but for now that's
      // how it is
      local
	height = AITile.GetHeight(foundtile);
      adjacent = AITileList();
      adjacent.AddTile(foundtile - AIMap.GetTileIndex(1, 0));
      adjacent.AddTile(foundtile - AIMap.GetTileIndex(0, 1));
      adjacent.AddTile(foundtile - AIMap.GetTileIndex(-1, 0));
      adjacent.AddTile(foundtile - AIMap.GetTileIndex(0, -1));

      adjacent.Valuate(AIMap.IsValidTile);
      adjacent.KeepValue(1);

      adjacent.Valuate(AITile.GetHeight);
      adjacent.KeepValue(height);
      adjacent.Valuate(AITile.GetSlope);
      adjacent.KeepValue(0);
      adjacent.Valuate(AITile.IsBuildable);
      adjacent.KeepValue(1);

      /* AITile.IsBuildable returns 0 for tiles with roads
         this is not appropriate behaviour for stations
         to compensate, this splits adjacent into two lists, run IsBuildable on one,
         checks the other for existing road tiles. drops 0 from each and merges

         for a depot we only want open tiles, so this does not need to be done */

      if (siteType >> 0)
	{
	  local
	    adjacent2 = AIList();
	  adjacent2.AddList(adjacent);
	  adjacent2.Valuate(AIRoad.IsRoadTile);
	  adjacent2.KeepValue(1);
	  adjacent.AddList(adjacent2);
	}
      count = adjacent.Count();
      // AILog.Info("Line: " + adjacent.Count() + " sites.");


    }

  adjacent.Valuate(AIBase.RandItem);
  adjacent.Sort(AIList.SORT_BY_VALUE, true);
  adjacent.KeepTop(1);
  local
    foundtile2 = adjacent.Begin();
  if (siteType >> 0)
    {
      return[foundtile, foundtile2];
    }
  else
    {
      return[foundtile2, foundtile];
    }
}






/*
 * function - Line.GetCargoRating()
 * a wrapper for AIStation.GetCargoRating
 *
 * returns cargo rating at station1
 */

function Line::GetCargoRating()
{
  local
    rating = AIStation.GetCargoRating(this.station1, this.cargo);
  return rating;
}





/*
 * function - Line.AddAdditionalVehicle()
 * adds a new vehicle to the line and shares orders with the first
 * also checks if this means another station should be called, and if so
 * calls ExpandStations.
 *
 * returns true if vehicle was added successfully
 */

function Line::AddAdditionalVehicle()
{
  local
    possibleEngines = AIEngineList(AIVehicle.VEHICLE_ROAD);
  possibleEngines.Valuate(AIEngine.GetCargoType);
  possibleEngines.KeepValue(this.cargo);
  possibleEngines.Valuate(AIEngine.GetReliability);
  possibleEngines.Sort(AIList.SORT_BY_VALUE, false);
  possibleEngines.KeepTop(1);
  local
    engineID = possibleEngines.Begin();

  // build the vehicle
  local
    thisvehicle = AIVehicle.BuildVehicle(this.depottile, engineID);
  if (!AIVehicle.IsValidVehicle(thisvehicle))
    {
      AILog.Warning("AddAdditionalVehicle: Error building vehicle at " +
		    this.depottile);
      return false;
    }

  // give it some orders
  if (!AIOrder.ShareOrders(thisvehicle, this.Vehicles[0]))
    {
      AILog.Warning("AddAdditionalVehicle: error giving orders to vehicle " +
		    thisvehicle);
      return false;
    }

  // and start it
  if (!AIVehicle.StartStopVehicle(thisvehicle))
    {
      AILog.Warning("AddAdditionalVehicle: error starting vehicle " +
		    thisvehicle);
      return false;
    }

  this.Vehicles.append(thisvehicle);
  local
    vehicles_per_station = this.Vehicles.len() / this.station_site_count;
  // AILog.Info("AddAdditionalVehicle: " + vehicles_per_station +
  //      " vehicles/station on this line.");

  if (vehicles_per_station > this.vps_target 
	&& (!this.station1_crowded || !this.station2_crowded) 
	&& robotAI.HasMoney(5000))
    {
      AILog.Info("AddAdditionalVehicle: Expanding stations on this line.");
      if (!this.ExpandStations())
	{
	  AILog.Warning("AddAdditionalVehicle: Could not expand stations!");
	}
    }

	if (this.group_id != -1)
	{
		AIGroup.MoveVehicle(this.group_id, thisvehicle);
	}
  return true;
}

/*
 *
 * function Line.ExpandStation(number)
 * adds an additional building to given station
 * 1 is at producer, 2 is at buyer
 *
 * returns true if the expansion was done successfuly
 *
 */

function
Line::ExpandStation(number)
{
  local site =[];
  local station_tiles;
  if (number == 1)
    station_tiles = AITileList_StationType(this.station1,
					   AIStation.STATION_TRUCK_STOP);
  if (number == 2)
    station_tiles = AITileList_StationType(this.station2,
					   AIStation.STATION_TRUCK_STOP);
  local stationtile;
  if (number == 1)
    stationtile = this.station1tile;
  if (number == 2)
    stationtile = this.station2tile;
  station_tiles.Valuate(AIRoad.IsRoadStationTile);
  station_tiles.KeepValue(1);
  station_tiles.Valuate(AITile.GetOwner);
  station_tiles.KeepValue(AICompany.ResolveCompanyID(AICompany.MY_COMPANY));

  if (station_tiles.IsEmpty())
    {

      AILog.Warning("ExpandStation: There's no station here");
      return false;

    }
  else
    {
      AILog.Info("ExpandStation: Found " + station_tiles.Count() +
		 " existing station sites.");
      // build a list of tiles that are adjacent to existing stations
      local candidate_sites = AIList();

      foreach(thistile, val in station_tiles)
      {

	local templist = AITileList();
	templist.AddTile(thistile - AIMap.GetTileIndex(-1, -1));
	templist.AddTile(thistile - AIMap.GetTileIndex(-1, 0));
	templist.AddTile(thistile - AIMap.GetTileIndex(-1, 1));
	templist.AddTile(thistile - AIMap.GetTileIndex(0, -1));
	templist.AddTile(thistile - AIMap.GetTileIndex(0, 1));
	templist.AddTile(thistile - AIMap.GetTileIndex(1, -1));
	templist.AddTile(thistile - AIMap.GetTileIndex(1, 0));
	templist.AddTile(thistile - AIMap.GetTileIndex(1, 1));

	templist.Valuate(AIMap.IsValidTile);
	templist.KeepValue(1);
	templist.Valuate(AITile.GetSlope);
	templist.KeepValue(0);
	templist.Valuate(AITile.IsBuildable);
	templist.KeepValue(1);

	/*
	   AILog.Info("ExpandStation: " + templist.Count() + 
	   " flat buildable tiles adjacent to station at " + thistile);
	 */

	candidate_sites.AddList(templist);
      }

      AILog.Info("ExpandStation: " + candidate_sites.Count() +
		 " candidate sites for station extension");

      if (candidate_sites.IsEmpty())
	{

	  // no places to extend the station
	  // so act like there's no station there at all
	  AILog.Info("ExpandStation: No candidate site, aborting.");
	  return false;
	}
      else
	{

	  // otherwise try to figure out which tile is best
	  local sites_with_roads = AIList();
	  sites_with_roads.AddList(candidate_sites);
	  sites_with_roads.Valuate(this.FindFacingRoadTile);
	  sites_with_roads.RemoveValue(-1);
	  local sites_are_empty = AIList();
	  sites_are_empty.AddList(candidate_sites);
	  sites_are_empty.Valuate(this.FindFacingEmptyTile);
	  sites_are_empty.RemoveValue(-1);

	  if (!sites_with_roads.IsEmpty())
	    {

	      sites_with_roads.Valuate(AIBase.RandItem);
	      sites_with_roads.Sort(AIList.SORT_BY_VALUE, false);
	      local site_tile = sites_with_roads.Begin();
	      local site_front = this.FindFacingRoadTile(site_tile);
	      site =[site_tile, site_front];
	      AILog.Info("ExpandStation: " + site[0] +
			 " already faces a road, building there.");

	    }
	  else if (!sites_are_empty.IsEmpty())
	    {

	      // if no sites facing a road tile, try for empty tiles
	      sites_are_empty.Valuate(AIBase.RandItem);
	      sites_are_empty.Sort(AIList.SORT_BY_VALUE, false);
	      local site_tile = sites_are_empty.Begin();
	      local site_front = this.FindFacingEmptyTile(site_tile);
	      site =[site_tile, site_front];
	      AILog.Info("ExpandStation: " + site[0] +
			 " is an empty site, building there.");

	    }
	  else
	    {

	      // no valid sites to build the extension
	      // so act like there's no station there at all
	      AILog.
		Info
		("ExpandStation: No buildable candidate sites, aborting");
	      return false;

	    }
	}

    }

	local counter = 0;
  while (!AIRoad.BuildRoad(site[0], site[1]))
    {

      if (AIError.GetLastError() == AIError.ERR_VEHICLE_IN_THE_WAY)
	{
		if (counter == 0)
		{
			AILog.Info("ExpandStation: Road vehicle in way, retrying.");
			counter = 10;
		}
		counter--;
	  AIController.Sleep(1);

	}
      else
	{

	  AILog.Error("ExpandStation: Error building road at " + site[1]);
	  return false;

	}
    }

  if (!AIRoad.BuildRoadStation(site[0], site[1], true, false, true))
    {

      AILog.Error("ExpandStation: Error building station at " + site[0]);
      AITile.DemolishTile(site[0]);

      return false;
    }

  // make sure we're connected to the original station

  local pathfinder1 = RoadPathFinder();
  pathfinder1.cost.no_existing_road = pathfinder1.cost.max_cost;
  pathfinder1.InitializePath([site[1]],[stationtile]);
  local path1 = false;
  while (path1 == false)
    {
      path1 = pathfinder1.FindPath(100);
      AIController.Sleep(1);
    }
  if (path1 == null)
    {

      // no path, gotta build one
      // first gotta run the pathfinder again
      pathfinder1.cost.no_existing_road = 2000;
      pathfinder1.cost.bridge_per_tile = 1500;
      pathfinder1.cost.tunnel_per_tile = 1300;

      local path1 = false;
      pathfinder1.InitializePath([site[1]],[stationtile]);
      while (path1 == false)
	{

	  path1 = pathfinder1.FindPath(100);
	  AIController.Sleep(1);
	}

      while (path1 != null)
	{

	  local par1 = path1.GetParent();
	  if (path1.GetParent() != null)
	    {

	      if (AIMap.
		  DistanceManhattan(path1.GetTile(), par1.GetTile()) == 1)
		{

// it is possible part of the path is blocked, so if building a tile fails
// check for that

		  // AILog.Info("Line: Building road...");
			
			local counter = 0;
		  while (!AIRoad.BuildRoad(path1.GetTile(), par1.GetTile()))
		    {

		      if (AIError.GetLastError() ==
			  AIError.ERR_VEHICLE_IN_THE_WAY)
			{
				if (counter == 0)
				{
			 	 	AILog.Info
			    			("ExpandStation: Road vehicle in the way, retrying.");
					counter = 10;
				}
			  counter--;
			  AIController.Sleep(1);

			}

		      else if (AIError.GetLastError() ==
			       AIError.ERR_NOT_ENOUGH_CASH)

			{

			  AILog.
			    Warning
			    ("ExpandStation: Error: Not enough cash to finish building route.");
			  return false;

			}
// if that's not the case, maybe the tiles are already connected
		      else if (AIRoad.
			       AreRoadTilesConnected(path1.GetTile(),
						     par1.GetTile()))
			{
			  break;
			}
// if that's not the case, whine
		      else
			{
			  AILog.
			    Warning("ExpandStation: Error building route");
			  return false;
			}
		    }

		}
	      else
		{		// it's a bridge or a tunnel

		  if (!AIBridge.IsBridgeTile(path1.GetTile()) &&
		      !AITunnel.IsTunnelTile(path1.GetTile()))
		    {

		      if (AIRoad.IsRoadTile(path1.GetTile()))
			AITile.DemolishTile(path1.GetTile());
		      if (AITunnel.GetOtherTunnelEnd(path1.GetTile()) ==
			  par1.GetTile())
			{

			  AILog.Info("ExpandStation: Building tunnel...");

				local counter = 0;
			  while (!AITunnel.
				 BuildTunnel(AIVehicle.VEHICLE_ROAD,
					     path1.GetTile()))
			    {

			      if (AIError.GetLastError() ==
				  AIError.ERR_VEHICLE_IN_THE_WAY)
				{
					if (counter == 0)
					{
					  AILog.Info
				 	   ("ExpandStation: Road vehicle in the way, retrying.");
						counter = 10;
					}
				  counter--;
				  AIController.Sleep(1);

				}
			      else if (AIError.GetLastError() ==
				       AIError.ERR_NOT_ENOUGH_CASH)
				{

				  AILog.
				    Warning
				    ("ExpandStation: Error: Not enough cash to finish building route.");
				  return false;

				}
			      else
				{

				  AILog.
				    Warning
				    ("ExpandStation: Error building route");
				  return false;

				}
			    }
			}
		      else
			{

			  local
			    bridges =
			    AIBridgeList_Length(AIMap.
						DistanceManhattan(path1.
								  GetTile
								  (),
								  par1.
								  GetTile() +
								  1));
			  bridges.Valuate(AIBridge.GetMaxSpeed);
			  bridges.Sort(AIAbstractList.SORT_BY_VALUE, false);
			  AILog.Info("ExpandStation: building bridge...");

			local counter = 0;
			  while (!AIBridge.
				 BuildBridge(AIVehicle.VEHICLE_ROAD,
					     bridges.Begin(),
					     path1.GetTile(), par1.GetTile()))
			    {

			      if (AIError.GetLastError() ==
				  AIError.ERR_VEHICLE_IN_THE_WAY)
				{
					if (counter == 0)
					{
						AILog.Info
							("ExpandStation: Road vehicle in the way, retrying.");
						counter = 10;
					}

				  counter--;
				  AIController.Sleep(1);

				}
			      else if (AIError.GetLastError() ==
				       AIError.ERR_NOT_ENOUGH_CASH)
				{

				  AILog.
				    Warning
				    ("ExpandStation: Error: Not enough cash to finish building route.");
				  return false;

				}
			      else
				{

				  AILog.
				    Warning
				    ("ExpandStation: Error building route");
				  return false;

				}
			    }
			}
		    }
		}
	    }
	  path1 = par1;
	}



    }
  return true;
}

/*
 * function Line.ExpandStations()
 * adds an additional building to each station on the line
 *
 * returns true if the expansion was done successfully
 */

function
Line::ExpandStations()
{
	local success = true;

	if (this.station1_crowded == false)
	{
		if (this.ExpandStation(1))
		{
			AILog.Info("ExpandStations: New station built at producer");
			this.station_site_count += 1;
		}

		else
		{
			AILog.Warning
				("ExpandStations: Could not expand station at producer");
			this.station1_crowded = true;
			success = false;
		}
	}

	if (this.station2_crowded == false)
	{
		if (this.ExpandStation(2))
		{
			AILog.Info("ExpandStations: New station built at buyer");
			this.station_site_count += 1;
		}

		else
		{
			AILog.Warning
				("ExpandStations: Could not expand station at buyer");
			this.station2_crowded = true;
			success = false;
		}

	}

	return success;

}



/*
 * function Line.GetNumVehiclesUnprofitable()
 *
 * polls each vehicle for its profit last year
 * if it is below 0 it increments a counter
 * returns the counter
 */

function Line::GetNumVehiclesUnprofitable()
{
	local counter = 0;
	foreach(index, vehicle_id in this.Vehicles)
	{
		if (AIVehicle.GetAge(vehicle_id) >= 750)
		{
			if (AIVehicle.GetProfitLastYear(vehicle_id) < 0)
				counter += 1;
		}
	}
	return counter;
}



/*
 * function Line.GetListVehiclesUnprofitable()
 *
 * polls each vehicle for its profit last year
 * if it is below 0 adds it to a list
 * returns the list
 */

function Line::GetListVehiclesUnprofitable()
{
	local list = [];
	foreach(index, vehicle_id in this.Vehicles)
	{
		if (AIVehicle.GetAge(vehicle_id) >= 750)
		{
			if (AIVehicle.GetProfitLastYear(vehicle_id) < 0)
				list.append(vehicle_id);
		}
	}
	return list;
}

/*
 * function Line.CullVechiles()
 *
 * sends unprofitable vehicles on this line to depot
 * sells them as they get there
 */

function Line::CullVehicles()
{

	local unprofitable = this.GetListVehiclesUnprofitable();
	local senttodepot = array(unprofitable.len(), 0);
	local numbersenttodepot = 0;
	local counter = 0;
	while (numbersenttodepot < unprofitable.len())
	{
		foreach(index, vehicle_id in unprofitable)
		{
			if (senttodepot[index] == 0) 
			{
				if (!AIVehicle.SendVehicleToDepot(vehicle_id))
				{
					if (counter == 0)
					{
						AILog.Info("CullVehicles: error sending vehicle" 
							+ " to depot, retrying.");
						counter = 100;
					}
					counter--;
					AIController.Sleep(1);
				}
				else
				{
					senttodepot[index] = 1;
					numbersenttodepot++;
				}
			}
		}
	}

	while (unprofitable.len() != 0)
	{
		for (local i = 0; i < unprofitable.len(); i++)
		{
			if (AIVehicle.IsStoppedInDepot(unprofitable[i]))
			{
				if (AIVehicle.SellVehicle(unprofitable[i]))
				{
					this.SearchAndDestroy(unprofitable[i]);
					unprofitable.remove(i);
					i--;
					AILog.Info("CullVehicles: " + unprofitable.len()
						+ " unprofitable vehicles remain.");
				}
			}
		}
	}

	AILog.Info("CullVehicles: done selling vehicles.");	
	return;

}



/*
 * function Line.SearchAndDestroy(vehicle_id)
 *
 * finds the given vehicle id in Line.Vehicles and removes it
 * just to keep Line.CullVehicles a bit neater
 */


function Line::SearchAndDestroy(vehicle_id)
{
	for (local i = 0; i < this.Vehicles.len(); i++)
		if (this.Vehicles[i] == vehicle_id)
		{
			this.Vehicles.remove(i);
			return;
		}
	AILog.Error("SearchAndDestroy: could not find vehicle " + vehicle_id);
	return;
}
