/* 
 *
 * robotAI
 * Version 1.1
 * July 26 2008
 * Released under GNU GPL
 *
 * robotAI ships freight. with trucks
 *
 */

import("pathfinder.road", "RoadPathFinder", 3);
require("line.nut");


/*
 *
 * Class - robotAI
 *
 */

class robotAI extends AIController
{
	stop = false;
	name = null;
	myCargo = null;
	lines = [];
	cargoExclude = [];

	buildCash = 60000;		// cash on hand required for new route
	updateCash = 20000;		//   "      "      "         new vehicle
	ratingtarget = 60;		// station rating to maintain
	  // this is a tradeoff between efficient service
	  // and a truck conga line

	function Start();
	function SetCompanyName();
	function ChooseBestCargo(exclude);
	function CargoEstimator(cargo_id);
	function HasMoney(CashNeeded);
	function CullVehicles();
	function RepayLoan();

}





/*
 * function RobotAI.Start
 */

function robotAI::Start()
{
  /* Say hi */
	AILog.Info("INIT: I'm alive!");

	this.SetCompanyName();
	this.name = AICompany.GetCompanyName(AICompany.MY_COMPANY);
	AILog.Info("INIT: Name set: " + this.name);

  /* check if breakdowns are on and if so enable autorenew */

	if (AIGameSettings.GetValue("vehicle_breakdowns") != 0)
	{
		AILog.Info("INIT: Breakdowns are enabled, setting autorenew.");
		AICompany.SetAutoRenewStatus(true);
		AICompany.SetAutoRenewMonths(-6);
		AICompany.SetAutoRenewMoney(100000);
	}

	this.myCargo = this.ChooseBestCargo(cargoExclude);
	AILog.Info("INIT: Gonna go for cargo #" + myCargo);

	/* initialize main loop */
	/* set timing counters  */
	local timer = 0;
	local timerBuildRoute = timer;
	local timerCheckRoute = timer + 3;
	local timerCullVehicles = timer + 150;
	local timerRepayLoan = timer + 1;

		// this guy can be a long time
		// because it we wanna give a vehicle 2 years to ensure last month's
		// profit tells the real story


	/* build lists of industries producing our cargo */
	local Producers = AIIndustryList_CargoProducing(myCargo);
	AILog.Info("INIT: " + Producers.Count() + " producer(s) in the world");
	local ProducersNoservice = AIList();
	ProducersNoservice.AddList(Producers);
  	local ProducersBest = AIList();
	local ChosenProducer = null;
	local ChosenProducerProd = null;
	local ChosenProducerLocation = null;


	/* build lists of industries accepting our cargo */
	local Buyers = AIIndustryList_CargoAccepting(myCargo);
	local ChosenBuyer = 0;
	local ChosenBuyerLocation = null;
	AILog.Info("INIT: " + Buyers.Count() + " buyer(s) in the world");

	AILog.Info("INIT: Ready to rock...");
	while (!this.stop)
	{				/* main loop */

		while (timer >= timerBuildRoute)
		{			/* time to build a route! */

			timerBuildRoute = timer + 5;
			if (ProducersNoservice.IsEmpty())
			{		// handle being out of producers
				AILog.Info("INFO: No new producers to service");
				AILog.Info("INFO: Choosing new cargo");
				if (cargoExclude.len() != 0)
				{
					if (cargoExclude.top() != myCargo)
						cargoExclude.append(myCargo);
				}
				else
				{
					cargoExclude.append(myCargo);
				}

				local newCargo = this.ChooseBestCargo(cargoExclude);
				if (newCargo != null)
				{
					this.myCargo = newCargo;
					AILog.Info("INFO: New cargo is #" + myCargo);
					Producers = AIIndustryList_CargoProducing(myCargo);
					AILog.Info("INFO: " + Producers.Count() +
						" producer(s) in the world");
					ProducersNoservice = AIList();
					ProducersNoservice.AddList(Producers);
					ProducersBest = AIList();
					ChosenProducer = null;
					ChosenProducerProd = null;
					ChosenProducerLocation = null;
	
					Buyers = AIIndustryList_CargoAccepting(myCargo);
					ChosenBuyer = 0;
					ChosenBuyerLocation = null;
					AILog.Info("INFO: " + Buyers.Count() +
						" buyer(s) in the world");
				}
				else
				{
					AILog.Error("INFO: Cound not choose new cargo!");
				}
			}
			else
			{		/* let's get building */
				if (!this.HasMoney(buildCash))
				{	/* if we have the cash */
					AILog.Info("INFO: Not enough cash to add new route");
				}
				else
				{
			  /* first choose the 5 most productive producers and randomly
			     choose one of them */
	
					ProducersBest = AIList();
					ProducersBest.AddList(ProducersNoservice);
					ProducersBest.Valuate(
						AIIndustry.GetLastMonthProduction, myCargo);
					ProducersBest.Sort(AIList.SORT_BY_VALUE, false);
					ProducersBest.KeepTop(5);

					ProducersBest.Valuate(AIBase.RandItem);
					ProducersBest.Sort(AIList.SORT_BY_VALUE, false);

					ChosenProducer = ProducersBest.Begin();
					ChosenProducerProd =
						AIIndustry.GetLastMonthProduction(ChosenProducer,
						myCargo);

					if (ChosenProducerProd == 0)
					{
						/* make sure we don't get tricked by dead 
							producers! */
						AILog.Info("INFO: Production at " +
							AIIndustry.GetName(ChosenProducer) +
						" is 0, aborting.");

						ProducersBest.KeepTop(1);
						ProducersNoservice.RemoveList(ProducersBest);
						break;
					}

					ChosenProducerLocation =
						AIIndustry.GetLocation(ChosenProducer);
					AILog.Info("INFO: Building a route from " +
						AIIndustry.GetName(ChosenProducer) + " (" +
						ChosenProducerProd + " units produced)... ");

					 /* then choose the closest buyer */
					 /* something wonky is going on here */

					Buyers.Valuate(AIIndustry.GetDistanceManhattanToTile,
						ChosenProducerLocation);
					Buyers.Sort(AIList.SORT_BY_VALUE, true);

					if (Buyers.IsEmpty())
					{
						AILog.Info
			("INFO: No buyers for this cargo, removing from list.");
						ProducersNoservice = AIList();
						break;
					}

					ChosenBuyer = Buyers.Begin();

					while (ChosenBuyer == ChosenProducer 
						&& Buyers.HasNext())
					{
						AILog.Info
			("INFO: Producer and buyer can not be same industry!");
						ChosenBuyer = Buyers.Next();
					}

					if (ChosenBuyer == ChosenProducer)
					{
						AILog.Warning
							("INFO: I think i just got tricked by a bank.");
						ProducersBest.KeepTop(1);
						ProducersNoservice.RemoveList(ProducersBest);
						break;
					}

					ChosenBuyerLocation = AIIndustry.GetLocation
						(ChosenBuyer);
					AILog.Info("INFO: ... to " +
						AIIndustry.GetName(ChosenBuyer) +
						", distance: " + Buyers.GetValue(ChosenBuyer));

					if (!AIIndustry.IsCargoAccepted(ChosenBuyer, myCargo))
					{
						AILog.Error
			("INFO: Illegitimate destination, removing producer from list.");
						ProducersBest.KeepTop(1);
						ProducersNoservice.RemoveList(ProducersBest);
						break;
					}

					/* here is where we actually build the route ... */

					local line = null;
					line = Line(ChosenProducer, ChosenBuyer, myCargo);
	
					if (!line.InitializeRoute(lines.len()))
					{
						AILog.Warning("INFO: Error initializing route!");
					}
					else
					{
						AILog.Info
			("INFO: Route added successfully, adding to list...");
					lines.append(line);
					}

					/* remove this producer from the list of unserviced
						producers */
					ProducersBest.KeepTop(1);
					ProducersNoservice.RemoveList(ProducersBest);
				}

			}
	
			this.RepayLoan();

		}

		this.Sleep(1);

		while (timer >= timerCheckRoute)
		{
			/* see if any of our routes need some sprucing up */
			timerCheckRoute = timer + 2;
			if (this.HasMoney(updateCash))
			{

				local poorLines = AIList();
				foreach(index, thisline in lines)
				{
					local thisrating = thisline.GetCargoRating();

					if (thisrating <= this.ratingtarget 
						&& thisline.GetNumVehiclesUnprofitable() == 0)
						// don't want to add more vehicles to unprofitable
						// routes
					{
						poorLines.AddItem(index, thisrating);
					}
				}

				if (poorLines.Count() >> 0)
				{
					AILog.Info("INFO: " + poorLines.Count() +
					" lines need more trucks.");
					poorLines.Sort(AIList.SORT_BY_VALUE, true);
					local thisline = poorLines.Begin();
	
					while (this.HasMoney(updateCash) 
						&& poorLines.HasNext())
					{
						if (!lines[thisline].AddAdditionalVehicle())
						{
							AILog.Warning
			("INFO: Could not add new truck to line " + thisline);
						}
						else
						{
							AILog.Info("INFO: added new truck to line " +
								thisline);
						}

					this.Sleep(1);
					thisline = poorLines.Next();
					}
				}
			}

			this.RepayLoan();

		}

		while (timer >= timerCullVehicles)
		{
			timerCullVehicles += 15;
			this.CullVehicles();
			this.RepayLoan();
		}


		timer = timer + 1;
		this.Sleep(40 + AIBase.RandRange(11));
	}

}






/*
 * function robotAI.SetCompanyName
 */

function robotAI::SetCompanyName()
{
	if (!AICompany.SetCompanyName("robotAI"))
	{
		local
		i = 2;
		while (!AICompany.SetCompanyName("robotAI #" + i))
		{
			this.Sleep(1);
			i = i + 1;
		}
	}
}







/*
 * function robotAI.ChooseBestCargo(exclude)
 */

function robotAI::ChooseBestCargo(exclude)
{

	/* identify freight...  */
	local CargoList = AICargoList();
	CargoList.Valuate(AICargo.IsFreight);
	CargoList.KeepValue(1);

	/* at least for now we don't want things that go to towns. */

	CargoList.Valuate(AICargo.GetTownEffect);
	CargoList.KeepValue(AICargo.TE_NONE);

	/* ... and let us know which one they are ... */

	AILog.Info("ChooseBestCargo: " + CargoList.Count() + " cargos present");
	local ExcludedCargos = AIList();
	foreach(index, thisExcluded in exclude)
	{
		if (CargoList.HasItem(thisExcluded))
		{
			AILog.Info("ChooseBestCargo: Cargo #" + thisExcluded +
				" is excluded.");
			ExcludedCargos.AddItem(thisExcluded, 1);
		}
	}
	
	CargoList.RemoveList(ExcludedCargos);

	if (CargoList.IsEmpty())
	{
		AILog.Error("ChooseBestCargo: Unable to choose a cargo!");
	return null;
	}

	local Cargo = CargoList.Begin();
	while (CargoList.HasNext())
	{
		AILog.Info("ChooseBestCargo: Cargo " + Cargo + ": " +
			AICargo.GetCargoLabel(Cargo));
		Cargo = CargoList.Next();
	}

	CargoList.Valuate(this.CargoEstimator);
	CargoList.Sort(AIList.SORT_BY_VALUE, false);
	local BestCargo = CargoList.Begin();

	return BestCargo;

}



/*
 * function robotAI.CargoEstimator(cargo_id)
 * a valuator for robotAI.ChooseBestCargo
 * finds total production for a cargo, and multiplies it by the cargo revenue
 */

function robotAI::CargoEstimator(cargo_id)
{
	if (!AICargo.IsValidCargo(cargo_id))
	{
		AILog.Error("CargoEstimator: invalid cargo id!");
		return -1;
	}
	local industry_list = AIIndustryList_CargoProducing(cargo_id);
	if (industry_list.IsEmpty())
	{
		AILog.Info("CargoEstimator: no producers of cargo " + cargo_id);
		return 0;
	}
	local world_production = 0;
	local this_industry = industry_list.Begin();

	while (industry_list.HasNext())
	{
		world_production = world_production +
			AIIndustry.GetLastMonthProduction(this_industry, cargo_id);
		industry_list.Next();
	}

	local cargo_revenue = AICargo.GetCargoIncome(cargo_id, 100, 30);

	local estimate = world_production * cargo_revenue;
	AILog.Info("CargoEstimator: " + cargo_id + " has a value of " + estimate);
	return estimate;
}




/*
 * function robotAI.HasMoney(CashNeeded)
 */

function robotAI::HasMoney(CashNeeded)
{
	local CashOnHand = AICompany.GetBankBalance(AICompany.MY_COMPANY);
	// AILog.Info("HasMoney: " + CashOnHand);
	if (CashOnHand > CashNeeded)
	{
		return true;
	}
	AICompany.SetLoanAmount(AICompany.GetMaxLoanAmount());
	CashOnHand = AICompany.GetBankBalance(AICompany.MY_COMPANY);
	if (CashOnHand > CashNeeded)
	{
		return true;
	}
	else
	{
		// AILog.Info("HasMoney: not enough cash");
		return false;
	}
}



/*
 * function robotAI.CullVehicles
 *
 * sells unprofitable vehicles
 */

function robotAI::CullVehicles()
{
	local worstline = -1;
	local worstline_number = 0;

	foreach (index, thisline in lines)
	{
		local this_number = thisline.GetNumVehiclesUnprofitable();

		if (this_number != 0)
			AILog.Info("CullVehicles: line " + index + ": " + this_number);

		if (this_number > worstline_number)
		{
			worstline = index;
			worstline_number = this_number;
		}
	}

	if (worstline != -1)
	{
		AILog.Info("CullVehicles: line " + worstline 
			+ " has the most unprofitable vehicles");
		lines[worstline].CullVehicles();
	}
	else
	{
		AILog.Info("CullVehicles: no unprofitable vehicles! very nice");
	}

	return true;
}



/*
 * function robotAI.RepayLoan()
 */

function robotAI::RepayLoan()
{
	local currentloan = AICompany.GetLoanAmount();
	// AILog.Info("RepayLoan: old loan is " + currentloan);
	local currentbalance = AICompany.GetBankBalance(AICompany.MY_COMPANY);
	local loaninterval = AICompany.GetLoanInterval();

	local minloanpossible = currentloan - currentbalance;
	local value = minloanpossible / loaninterval;
	local newloan = value.tointeger() * loaninterval

	while (!AICompany.SetLoanAmount(newloan))
	{
		newloan += loaninterval;
	}

	// AILog.Info("RepayLoan: new loan is " + (newloan));
	return;
}
